/*
 *  Copyright (C) 2012 Khaldoon Ghanem
 *  See the file license.txt for copying permission.
 */

#include <stdio.h>
#include <stdlib.h>


#if defined(WIN32)
	#define GLEW_STATIC
	#include <GL/glew.h>
	#include <windows.h>
#else
	#include <GL/glew.h>
#endif


#if defined(__APPLE__) || defined(MACOSX)
	#include <OpenGL/gl.h>
	#include <OpenGL/glu.h>
	#include <GLUT/glut.h>
#else
	#include <GL/glut.h>
	#include <GL/gl.h>
	#include <GL/glu.h>
#endif


#include <math.h>
#include <string.h>

float pi;
float r, theta, phi;

GLuint v, f, p, loc;
//int l, m;
#define MAX_TERMS 32 
int count;
float data [MAX_TERMS][4];

void init(void)
{
    GLfloat mat_ambient[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat mat_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };
    GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat mat_shininess[] = { 100.0 };
    GLfloat white_light[] = { 1.0, 1.0, 1.0, 1.0 };
    GLfloat lmodel_ambient[] = { 0.1, 0.1, 0.1, 1.0 };
    GLfloat light_position[] = { 0.0, 0.0, 1.0, 0.0 };
    
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glShadeModel (GL_SMOOTH);

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);

    glLightfv(GL_LIGHT0, GL_DIFFUSE, white_light);
    glLightfv(GL_LIGHT0, GL_SPECULAR, white_light);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);

    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glEnable (GL_DEPTH_TEST);
    //glEnable(GL_LIGHTING);
    //glEnable(GL_LIGHT0);

}

void drawGrid(float spacing, float min, float max)
{
    float coord;
    glLineWidth(1);
    glBegin (GL_LINES);
    //Draw X-Y Plane            
    glColor3f(0.2, 0.2, 0.2);
    coord = spacing;
    while (coord <= max)
    {
        glVertex3f(coord, min, 0);
        glVertex3f(coord, max, 0);
        glVertex3f(min, coord, 0);
        glVertex3f(max, coord, 0);
        coord += spacing;
    }
    coord = -spacing;
    while (coord >= min)
    {
        glVertex3f(coord, min, 0);
        glVertex3f(coord, max, 0);
        glVertex3f(min, coord, 0);
        glVertex3f(max, coord, 0);
        coord -= spacing;
    }
    glEnd();
    glLineWidth(3);
    glBegin(GL_LINES);
    //Draw Axes    

    glColor3f(0.3, 0, 0);
    glVertex3f(min, 0, 0);
    glVertex3f(0, 0, 0);
    glColor3f(1, 0, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(max, 0, 0);

    glColor3f(0, 0.3, 0);
    glVertex3f(0, min, 0);
    glVertex3f(0, 0, 0);
    glColor3f(0, 1, 0);
    glVertex3f(0, 0, 0);
    glVertex3f(0, max, 0);

    glColor3f(0, 0, 0.3);
    glVertex3f(0, 0, min);
    glVertex3f(0, 0, 0);
    glColor3f(0, 0, 1);
    glVertex3f(0, 0, 0);
    glVertex3f(0, 0, max);

    glEnd();
}


void display(void)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    //
    glLoadIdentity();
    float x = r * sin(theta) * cos(phi);
    float y = r * sin(theta) * sin(phi);
    float z = r * cos(theta);

    gluLookAt(x, y, z, 0.0, 0.0, 0.0, 0.0f, 0.0f, 1.0f);

    //Draw grid
    glUseProgram(0);
    drawGrid(0.1, -10, 10);

    //Draw harmonics
    glUseProgram(p);
    
    loc = glGetUniformLocation(p, "data");
    glUniform4fv(loc, count, (float*)data);
    
    loc = glGetUniformLocation(p, "count");
    glUniform1i(loc, count);    

    glutSolidSphere(1.0, 256, 256);
    glutSwapBuffers();
}

int gw, gh;
void reshape(int w, int h)
{
    // Prevent a divide by zero, when window is too short
    // (you cant make a window of zero width).
    if (h == 0)
        h = 1;

    float ratio = 1.0 * w / h;

    // Reset the coordinate system before modifying
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();

    // Set the viewport to be the entire window
    glViewport(0, 0, w, h);

    // Set the correct perspective.
    gluPerspective(45, ratio, 0.1, 100);
    glMatrixMode (GL_MODELVIEW);
    gh = h;
    gw = w;
}

int fullscreen = 0;
int screen_pos_x;
int screen_pos_y;
int screen_w;
int screen_h;
void processNormalKeys(unsigned char key, int x, int y)
{
    //Navigation
    if (key == 'w' || key == 'W')
    {
        if (theta - 0.1 > 0)
            theta -= 0.1;
    }
    if (key == 's' || key == 'S')
    {
        if (theta + 0.1 < pi)
            theta += 0.1;
    }

    if (key == 'a' || key == 'A')
    {
        phi -= 0.1;
    }
    if (key == 'd' || key == 'D')
    {
        phi += 0.1;
    }

    if (key == 'q' || key == 'Q')
    {
        r = r * 1.2;
    }

    if (key == 'e' || key == 'E')
    {
        r = r / 1.2;
    }

    if (key == 'p' || key == 'P')
    {
        printf("Camera Position:");
        printf("r = %lg, theta = %lg, phi= %lg ", r, theta, phi);
        float x = r * sin(theta) * cos(phi);
        float y = r * sin(theta) * sin(phi);
        float z = r * cos(theta);
        printf("(x = %lg, y = %lg, z= %lg).\n", x, y, z);
    }

    if (key == 'f' || key == 'F')
    {
        fullscreen = !fullscreen;
        if (fullscreen)
        {
            screen_pos_x = glutGet((GLenum) GLUT_WINDOW_X);
            screen_pos_y = glutGet((GLenum) GLUT_WINDOW_Y);
            screen_w = gw;
            screen_h = gh;
            glutFullScreen();
        }
        else
        {
            glutReshapeWindow(screen_w, screen_h);
            glutPositionWindow(screen_pos_x, screen_pos_y);
        }
    }
    if (key == 27)
        exit(0);
}

float mouse_x = 0;
float mouse_y = 0;
float old_theta = 0, old_phi = 0;
int pressed = 0;
void mouseButton(int button, int state, int x, int y)
{
    if (button == GLUT_LEFT_BUTTON)
    {

        // when the button is released
        if (state == GLUT_UP)
        {
            pressed = 0;
        }
        else
        { // state = GLUT_DOWN
            mouse_x = x;
            mouse_y = y;
            old_theta = theta;
            old_phi = phi;
            pressed = 1;
        }
    }
    // Wheel reports as button 3(scroll up) and button 4(scroll down)
    if ((button == 3 && state == GLUT_DOWN)
            || (button == 4 && state == GLUT_DOWN))
    {
        float zoom;
        if (button == 4)
            r *= 1.2;
        else
            r /= 1.2;
    }
}

void mouseMove(int x, int y)
{
    if (pressed)
    {
        theta = old_theta + (float) (mouse_y - y) / gh * pi;
        if (theta < 0)
            theta = 0.001;
        if (theta > pi)
            theta = pi - 0.001;
        phi = old_phi + (float) (mouse_x - x) / gw * pi;
    }
}


char *textFileRead(char *fn) {


	FILE *fp;
	char *content = NULL;

	int count=0;

	if (fn != NULL) {
		fp = fopen(fn,"rt");

		if (fp != NULL) {
      
      fseek(fp, 0, SEEK_END);
      count = ftell(fp);
      rewind(fp);

			if (count > 0) {
				content = (char *)malloc(sizeof(char) * (count+1));
				count = fread(content,sizeof(char),count,fp);
				content[count] = '\0';
			}
			fclose(fp);
		}
	}
	return content;
}


#define printOpenGLError() printOglError(__FILE__, __LINE__)

int printOglError(char *file, int line)

{
    //
    // Returns 1 if an OpenGL error occurred, 0 otherwise.
    //

    GLenum glErr;
    int retCode = 0;

    glErr = glGetError();
    while (glErr != GL_NO_ERROR)

    {
        printf("glError in file %s @ line %d: %s\n", file, line,
                gluErrorString(glErr));
        retCode = 1;
        glErr = glGetError();
    }

    return retCode;
}

void printShaderInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten = 0;
    char *infoLog;

    glGetShaderiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *) malloc(infologLength);
        glGetShaderInfoLog(obj, infologLength, &charsWritten, infoLog);
        if (strlen(infoLog) > 0)
            printf("%s\n", infoLog);
        free(infoLog);
    }
}

void printProgramInfoLog(GLuint obj)
{
    int infologLength = 0;
    int charsWritten = 0;
    char *infoLog;

    glGetProgramiv(obj, GL_INFO_LOG_LENGTH, &infologLength);

    if (infologLength > 0)
    {
        infoLog = (char *) malloc(infologLength);
        glGetProgramInfoLog(obj, infologLength, &charsWritten, infoLog);
        if (strlen(infoLog) > 0)
            printf("%s\n", infoLog);
        free(infoLog);
    }
}

void setShaders()
{

    char *vs = NULL, *fs = NULL;

    v = glCreateShader(GL_VERTEX_SHADER);
    f = glCreateShader(GL_FRAGMENT_SHADER);

    vs = textFileRead("harmonics.vert");
    fs = textFileRead("harmonics.frag");

    const char * vv = vs;

    const char * ff = fs;

    glShaderSource(v, 1, &vv, NULL);
    glShaderSource(f, 1, &ff, NULL);

    free(vs);
    free(fs);

    glCompileShader(v);
    int isCompiled;
    glGetShaderiv(v, GL_COMPILE_STATUS, &isCompiled);
    if(isCompiled == 0)
    {
        printf("Vertex Shader - Compilation Error!\n");
        printShaderInfoLog(v);
        exit(1);
    }
        
    glCompileShader(f);
    glGetShaderiv(f, GL_COMPILE_STATUS, &isCompiled);
    if(isCompiled == 0)
    {
        printf("Fragment Shader - Compilation Error!\n");
        printShaderInfoLog(f);
        exit(1);
    }
    
    p = glCreateProgram();

    glAttachShader(p, v);
    glAttachShader(p, f);

    glLinkProgram(p);
    printProgramInfoLog(p);
    int isLinked;
    glGetProgramiv(p, GL_LINK_STATUS, (int *)&isLinked);
    if(isLinked == 0)
    {
        printf("Shading Program - Linking Error!\n");
        printProgramInfoLog(p);
        exit(1);
    }

}

#define LINE_WIDTH 256
void readData()
{

    count = 0;
    char line [LINE_WIDTH+1];
    if(fgets(line, LINE_WIDTH, stdin)==NULL&&!feof(stdin))
    {       
        printf("error: cannot read input stream.\n");
        exit(1);
    }
    
    while(!feof(stdin))
    {                        
        if(count==MAX_TERMS)
        {
            printf("error: Maximum number of terms is %d.\n", MAX_TERMS);
            exit(1);
        }
        
        int l,m;
        if(sscanf(line, " %d %d %f %f ", &l, &m, 
                &data[count][2], &data[count][3])!=4)
        {
            printf("line %d: error: cannot parse input line.\n", count+1);
            exit(1);
        }
        if (l < 0)
        {
            printf("line %d: error: l should be greater than zero.\n", count+1);
            exit(1);
        }
    
        if (m > l || (m < -1 * l))
        {
            printf("line: %d: error: m should be between l and -l.\n", count+1);
            exit(1);
        }
        data[count][0] = l;
        data[count][1] = m;
        count++;
        
        if(fgets(line, LINE_WIDTH, stdin)==NULL&&!feof(stdin))
        {       
            printf("error: cannot read input stream.\n");
            exit(1);
        }        
    }
    if(count==0)
    {
        printf("error: At least one term should be specified.\n");
        exit(1);
    }    
}

int main(int argc, char** argv)
{
    pi = atan(1)*4;
    r = 2;
    theta = pi / 2.5;
    phi = pi / 4.0;
    
    readData();
    
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(800, 600);
    glutInitWindowPosition(0, 0);
    glutCreateWindow("Spherical Harmonics");

    glutDisplayFunc(display);
    glutIdleFunc(display);    
    glutReshapeFunc(reshape);
    glutKeyboardFunc(processNormalKeys);
    glutMouseFunc(mouseButton);
    glutMotionFunc(mouseMove);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0"))
    {
        printf("OpenGL 2.0 not supported\n");
        exit(1);
    }

    setShaders();
    init();
    
    char* control = textFileRead("control.txt");
    printf("%s", control);
    free(control);
    
    glutMainLoop();
    return 0;
}

